import sys
from inspect import cleandoc

import click

from note_storage.cli.commands import drop
from note_storage.cli.commands.edit import update_note
from note_storage.cli.errors import exit_on_unsuccesful_response
from note_storage.cli.utils import dumps_note, loads_note
from note_storage.cli.utils.output import MINIMAL_NOTE_LINES, decorate_note
from note_storage.gateways import SQLiteNoteGateway
from note_storage.interactors import (
    DeleteLastReviewedNoteIdInteractor,
    GetAllNoteIDsInteractor,
    GetLastReviewedNoteIdInteractor,
    GetNotesWithTagsByFieldInteractor,
    SetLastReviewedNoteIdInteractor,
)
from note_storage.requests import (
    DeleteLastReviewedNoteIdRequest,
    GetAllNoteIDsRequest,
    GetLastReviewedNoteIdRequest,
    GetNotesWithTagsByFieldRequest,
    SetLastReviewedNoteIdRequest,
)

HELP_TEXT = cleandoc(
    """
    s - skip note
    e - edit note
    d - delete note
    q - quit review mode
    ? - print this message
    """
)


def get_current_note_action(note_number, all_notes_number):
    review_progress = int(note_number / all_notes_number * 100)

    return click.prompt(
        text='({0}/{1}, {2}%) Choose an action'.format(
            note_number, all_notes_number, review_progress
        ),
        type=click.Choice(['s', 'e', 'd', 'q', '?']),
        default='?',
    )


def print_review_help():
    click.secho(HELP_TEXT, fg='red', bold=True)


async def get_note_with_tags(db, note_id):
    interactor = GetNotesWithTagsByFieldInteractor(SQLiteNoteGateway(db))

    response = await interactor.execute(
        await GetNotesWithTagsByFieldRequest.from_dict(
            {'field': 'id', 'values': [note_id]}
        )
    )

    exit_on_unsuccesful_response(response)

    return response.value[0]


def print_note(note, tags):
    _, terminal_lines = click.get_terminal_size()
    max_body_lines = terminal_lines - MINIMAL_NOTE_LINES - 1

    click.echo(
        decorate_note(note, tags, query=[], max_body_lines=max_body_lines)
    )


async def get_all_notes_ids(db):
    note_gateway = SQLiteNoteGateway(db)
    interactor = GetAllNoteIDsInteractor(note_gateway)
    response = await interactor.execute(
        await GetAllNoteIDsRequest.from_dict({})
    )

    exit_on_unsuccesful_response(response)

    return response.value


async def edit_note(db, note, tags):
    tag_names = [tag.name for tag in tags]
    text = click.edit(
        dumps_note(note.body, tag_names), extension=('.' + note.extension)
    )
    if not text:
        sys.exit('Editing aborted')

    await update_note(db, note.id, note.extension, *loads_note(text))


async def get_last_reviewed_note_id(db):
    gateway = SQLiteNoteGateway(db)
    interactor = GetLastReviewedNoteIdInteractor(gateway)
    response = await interactor.execute(
        await GetLastReviewedNoteIdRequest.from_dict({})
    )

    exit_on_unsuccesful_response(response)

    return response.value


async def set_last_reviewed_note_id(db, last_reviewed_id):
    gateway = SQLiteNoteGateway(db)
    interactor = SetLastReviewedNoteIdInteractor(gateway)
    response = await interactor.execute(
        await SetLastReviewedNoteIdRequest.from_dict(
            {'last_reviewed_id': last_reviewed_id}
        )
    )

    exit_on_unsuccesful_response(response)

    return response.value


async def delete_last_reviewed_note_id(db):
    gateway = SQLiteNoteGateway(db)
    interactor = DeleteLastReviewedNoteIdInteractor(gateway)
    response = await interactor.execute(
        await DeleteLastReviewedNoteIdRequest.from_dict({})
    )

    exit_on_unsuccesful_response(response)

    return response.value


def find_nearest_greater_index(haystack, needle):
    for index, value in enumerate(haystack):
        if value > needle:
            return index
    return -1


def resume_review_prompt(last_reviewed_id, note_ids):
    CONTINUE_REVIEW_TEXT = 'Continue last review or start new?'
    DELETED_NOTE_RESUME_TEXT = (
        'Last reviewed note was deleted. '
        'Continue review from next note or start new?'
    )

    resume_choices = click.Choice(['c', 'n'])

    if last_reviewed_id in note_ids:
        if click.prompt(CONTINUE_REVIEW_TEXT, type=resume_choices) == 'c':
            note_ids = note_ids[note_ids.index(last_reviewed_id) :]
    else:
        if click.prompt(DELETED_NOTE_RESUME_TEXT, type=resume_choices) == 'c':
            note_ids = note_ids[
                find_nearest_greater_index(note_ids, last_reviewed_id) :
            ]

    return note_ids


async def review_note(db, note, tags, note_number, all_notes_number):
    while True:
        print_note(note, tags)

        action = get_current_note_action(note_number, all_notes_number)
        while action == '?':
            print_review_help()
            action = get_current_note_action(note_number, all_notes_number)

        if action == 's':
            break

        if action == 'q':
            sys.exit(0)

        if action == 'e':
            await edit_note(db, note, tags)
            break

        if action == 'd':
            if click.confirm('Are you sure?'):
                await drop(db, note.hex_id)
                break
