import asyncio
import logging
import pathlib
import random
import sys
from glob import glob
from importlib.metadata import version

import click

from note_storage.cli import strings
from note_storage.cli.commands import (
    add_tags,
    drop,
    push,
    remove_tags,
    rename_tag,
    search,
)
from note_storage.cli.commands.completions import complete_tag_name
from note_storage.cli.commands.edit import get_note_by_hex, update_note
from note_storage.cli.commands.full_review import (
    delete_last_reviewed_note_id,
    get_all_notes_ids,
    get_last_reviewed_note_id,
    get_note_with_tags,
    resume_review_prompt,
    review_note,
    set_last_reviewed_note_id,
)
from note_storage.cli.commands.search import get_hex_ids_by_note_ids
from note_storage.cli.utils import (
    copy_to_clipboard,
    db_context,
    dumps_note,
    loads_note,
    print_note_by_hex_id,
    print_notes_by_ids,
)
from note_storage.core.config import config
from note_storage.core.constants import PROJECT_NAME

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(
    version(PROJECT_NAME), message=f'{PROJECT_NAME} %(version)s'
)
def main():
    """Helps you to store a huge amount of notes

    <https://github.com/lxmnk/note-storage>
    """

    logging.getLogger('aiosqlite').disabled = True


def complete_path(ctx, args, incomplete):
    # pylint: disable=unused-argument
    completions = [
        x for x in glob(incomplete + '*') if x.startswith(incomplete)
    ]
    if len(completions) != 1:
        return completions

    path = pathlib.Path(completions[0])
    if not path.exists() or not path.is_dir():
        return completions

    return glob(completions[0] + '/*')


def complete_tag(ctx, args, incomplete):
    # pylint: disable=unused-argument
    async def wrapper():
        async with db_context() as db:
            return await complete_tag_name(db, incomplete, tag_names)

    tag_names = args[1:]

    if not (tag_names or incomplete):
        return []  # don't autocomplete empty query

    if tag_names and tag_names[-1] == incomplete:
        tag_names.pop()  # fish-shell doubles incomplete to args

    double_dash_index = tag_names.index('--') if '--' in tag_names else -1
    if double_dash_index != -1:
        tag_names = tag_names[double_dash_index + 1 :]

    return asyncio.run(wrapper())


@main.command('add-tags')
@click.argument('hex_id', metavar='ID', required=True)
@click.argument('tags', metavar='TAGS', required=True)
@click.option('-q', '--quiet', is_flag=True, help=strings.QUIET_OPTION_HELP)
def add_tags_command(quiet, hex_id, tags):
    """Add tags to note.

    TAGS should be space-separated list of tags surrounded by quotes.
    If you want to add one tag, quotes can be ommited.
    """

    async def wrapper():
        async with db_context() as db:
            note, current_tags = await get_note_by_hex(db, hex_id)
            current_tag_names = {tag.name for tag in current_tags}

            new_tags = list(set(tags_to_add) - current_tag_names)
            await add_tags(db, note.id, new_tags)

            if already_added_tags := current_tag_names & set(tags_to_add):
                formatted_tag_names = [
                    click.style(tag, bold=True) for tag in already_added_tags
                ]

                click.echo(
                    'Warning: note already has {0} tag(s)'.format(
                        ', '.join(formatted_tag_names)
                    )
                )

    tags_to_add = tags.strip().split()

    asyncio.run(wrapper())

    if not quiet:
        click.echo(
            'Tags was successfully added to note {}'.format(
                click.style(hex_id, bold=True)
            )
        )


@main.command('dump')
@click.argument('hex_id', metavar='ID', required=True)
@click.option(
    '-t', '--with-tags', is_flag=True, help=strings.WITH_TAGS_OPTION_HELP
)
@click.option(
    '-c', '--to-clipboard', is_flag=True, help=strings.TO_CLIPBOARD_OPTION_HELP
)
def dump_command(to_clipboard, with_tags, hex_id):
    """Get note content.

    Useful to sharing note content or passing it to another program.
    """

    async def wrapper():
        async with db_context() as db:
            note, tags = await get_note_by_hex(db, hex_id)
            tag_names = [tag.name for tag in tags] if with_tags else []

            return dumps_note(note.body, tag_names)

    content = asyncio.run(wrapper())

    if to_clipboard:
        copy_to_clipboard(content)
        click.echo(
            'Note '
            + click.style(hex_id, bold=True)
            + ' was copied to clipboard'
        )
        return

    click.echo(content)


@main.command('edit')
@click.option(
    '-e', '--extension', metavar='EXT', help=strings.EXTENSION_OPTION_HELP
)
@click.argument('hex_id', metavar='ID', required=True)
def edit_command(hex_id, extension):
    """Edit note with text editor.

    Preffered text editor should be assigned to $VISUAL or $EDITOR.
    """

    async def wrapper():
        async with db_context() as db:
            note, tags = await get_note_by_hex(db, hex_id)
            tag_names = [tag.name for tag in tags]
            note_ext = extension or note.extension

            text = click.edit(
                dumps_note(note.body, tag_names), extension=('.' + note_ext)
            )
            if not text:
                sys.exit('Editing aborted')

            await update_note(db, note.id, note_ext, *loads_note(text))

            if config.get('cli', 'instant_view'):
                await print_note_by_hex_id(db, hex_id)
            else:
                click.echo('Note ' + click.style(hex_id, bold=True) + ' saved')

    asyncio.run(wrapper())


@main.command('full-review')
def full_review_command():
    """Perform review of all notes.

    Helps to keep your notes up to date.
    """

    async def wrapper():
        async with db_context() as db:
            last_reviewed_id = await get_last_reviewed_note_id(db)
            note_ids = await get_all_notes_ids(db)

            if last_reviewed_id:
                note_ids = resume_review_prompt(last_reviewed_id, note_ids)

            all_notes_number = len(note_ids)

            if not all_notes_number:
                click.echo("Empty storage, can't start review")
                sys.exit(1)

            for note_number, note_id in enumerate(note_ids, start=1):
                note, tags = await get_note_with_tags(db, note_id)

                await set_last_reviewed_note_id(db, note.id)

                await review_note(
                    db, note, tags, note_number, all_notes_number
                )

            await delete_last_reviewed_note_id(db)

    asyncio.run(wrapper())


@main.command('new')
@click.option(
    '-t',
    '--tags',
    metavar='TAGS',
    help=strings.TAGS_OPTION_HELP,
)
@click.option(
    '-e',
    '--extension',
    metavar='EXT',
    help=strings.EXTENSION_OPTION_HELP,
    default=config.get('general', 'default_extension'),
    show_default=True,
)
def new_command(extension, tags):
    """Create note with text editor.

    Preffered text editor should be assigned to $VISUAL or $EDITOR.
    """

    async def wrapper():
        async with db_context() as db:
            hex_id = await push(db, extension, note_body, tag_names)

            if config.get('cli', 'instant_view'):
                await print_note_by_hex_id(db, hex_id)
            else:
                click.echo(
                    'Note saved with id: ' + click.style(hex_id, bold=True)
                )

    text = click.edit(extension=('.' + extension))
    if not text:
        sys.exit('Empty note content')

    note_body, tag_names = (
        (text.strip(), tags.split()) if tags else loads_note(text)
    )

    asyncio.run(wrapper())


@main.command('pop')
@click.argument('hex_id', metavar='ID', required=True)
@click.option(
    '-c', '--to-clipboard', is_flag=True, help=strings.TO_CLIPBOARD_OPTION_HELP
)
@click.option(
    '-D', '--skip-dump', is_flag=True, help=strings.SKIP_DUMP_OPTION_HELP
)
@click.option(
    '-t', '--with-tags', is_flag=True, help=strings.WITH_TAGS_OPTION_HELP
)
def pop_command(with_tags, skip_dump, to_clipboard, hex_id):
    """Dump note to stdout and remove it from storage.

    Useful when moving notes to another app.
    """

    async def wrapper():
        async with db_context() as db:
            if not skip_dump:
                note, tags = await get_note_by_hex(db, hex_id)

                tag_names = [tag.name for tag in tags] if with_tags else []
                content = dumps_note(note.body, tag_names)

                if to_clipboard:
                    copy_to_clipboard(content)
                else:
                    click.echo(content)

            await drop(db, hex_id)

    if skip_dump and to_clipboard:
        sys.exit('"--skip-dump" can\'t be used with "--to-clipboard"')

    if skip_dump and with_tags:
        sys.exit('"--skip-dump" can\'t be used with "--with-tags"')

    asyncio.run(wrapper())

    if to_clipboard:
        click.echo(
            'Note '
            + click.style(hex_id, bold=True)
            + ' was copied to clipboard and removed'
        )

    if skip_dump:
        click.echo('Note ' + click.style(hex_id, bold=True) + ' removed')


@main.command('push')
@click.argument(
    'note_content',
    type=click.File('r'),
    required=False,
    autocompletion=complete_path,
)
@click.option('-k', '--keep', is_flag=True, help=strings.KEEP_OPTION_HELP)
@click.option('-q', '--quiet', is_flag=True, help=strings.ID_ONLY_OPTION_HELP)
@click.option(
    '-t',
    '--tags',
    metavar='TAGS',
    help=strings.TAGS_OPTION_HELP,
)
@click.option(
    '-e',
    '--extension',
    metavar='EXT',
    help=strings.EXTENSION_OPTION_HELP,
    default="txt",
    show_default=True,
)
def push_command(extension, tags, quiet, keep, note_content):
    """Move note content to storage.

    If NOTE_CONTENT is ommitted, `push` will read stdin.
    """

    async def wrapper():
        async with db_context() as db:
            hex_id = await push(db, extension, note_body, tag_names)

            if quiet:
                click.echo(hex_id)
                return

            if config.get('cli', 'instant_view'):
                await print_note_by_hex_id(db, hex_id)
            else:
                click.echo(
                    'Note saved with id: ' + click.style(hex_id, bold=True)
                )

    if not note_content and keep:
        sys.exit('"--keep" cannot be used with "stdin"')

    text = (
        note_content.read().strip()
        if note_content
        else click.get_text_stream('stdin').read().strip()
    )

    note_body, tag_names = (text, tags.split()) if tags else loads_note(text)

    asyncio.run(wrapper())

    if note_content and not keep:
        pathlib.Path(note_content.name).unlink()


@main.command('remove-tags')
@click.option('-q', '--quiet', is_flag=True, help=strings.QUIET_OPTION_HELP)
@click.argument('hex_id', metavar='ID', required=True)
@click.argument('tags', metavar='TAGS', required=True)
def remove_tags_command(quiet, hex_id, tags):
    """
    Remove tags from note.

    TAGS should be space-separated list of tags surrounded by quotes.
    If you want to remove one tag, quotes can be ommited.
    """

    async def wrapper():
        async with db_context() as db:
            note, current_tags = await get_note_by_hex(db, hex_id)
            current_tag_names = {tag.name for tag in current_tags}

            old_tags = list(current_tag_names & set(tags_to_remove))
            await remove_tags(db, note.id, old_tags)

            if missing_tags := set(tags_to_remove) - current_tag_names:
                formatted_tag_names = [
                    click.style(tag, bold=True) for tag in missing_tags
                ]

                click.echo(
                    "Warning: note doesn't have {0} tag(s)".format(
                        ', '.join(formatted_tag_names)
                    )
                )

    tags_to_remove = tags.strip().split()

    asyncio.run(wrapper())

    if not quiet:
        click.echo(
            'Tags was successfully removed from note {}'.format(
                click.style(hex_id, bold=True)
            )
        )


@main.command('rename-tag')
@click.option('-q', '--quiet', is_flag=True, help=strings.QUIET_OPTION_HELP)
@click.argument('old_name', metavar='OLD_NAME', autocompletion=complete_tag)
@click.argument('new_name', metavar='NEW_NAME')
def rename_command(new_name, old_name, quiet):
    """Change tag name.

    NEW_NAME should not be used by another tag.
    """

    async def wrapper():
        async with db_context() as db:
            return await rename_tag(db, old_name, new_name)

    asyncio.run(wrapper())

    if not quiet:
        click.echo(
            'Tag '
            + click.style(old_name, bold=True)
            + ' successfully renamed to '
            + click.style(new_name, bold=True)
        )


@main.command('search')
@click.argument(
    'tag_names',
    metavar='TAGS',
    nargs=-1,
    required=True,
    autocompletion=complete_tag,
)
@click.option(
    '-m',
    '--trim-at',
    metavar='N',
    type=int,
    help=strings.TRIM_OPTION_HELP,
    default=5,
    show_default=True,
)
@click.option(
    '-s', '--shuffle', is_flag=True, help=strings.SHUFFLE_OPTION_HELP
)
@click.option('-q', '--quiet', is_flag=True, help=strings.ID_ONLY_OPTION_HELP)
def search_command(quiet, shuffle, trim_at, tag_names):
    """Search note by tags.

    Autocompletion suggests tags located in notes together with already
    entered tags.

    Autocompletion doesn't work for an empty query. You should at least
    type the first letter of the first tag (other tags will be
    autocompleted without the first letter). It's like a search engine
    doesn't get you all pages if your search query is empty.

    If you use command options, you should separate ones from tags with
    -- (double-dash). Example:

    note search -m 8 -- tag1 tag2

    Otherwise any tag starting with dash(es) will be considered as
    option and autocompletion will be broken.
    """

    async def wrapper():
        async with db_context() as db:
            note_ids = await search(db, list(tag_names))

            if need_shuffle:
                random.shuffle(note_ids)

            if not note_ids:
                sys.exit('No notes found')

            if quiet:
                hex_ids = await get_hex_ids_by_note_ids(db, note_ids)
                click.echo('\n'.join(hex_ids))
                return

            return note_ids

    need_shuffle = config.get('general', 'shuffle_search') or shuffle
    if note_ids := asyncio.run(wrapper()):
        print_notes_by_ids(
            note_ids, tag_names, shuffle=need_shuffle, max_body_lines=trim_at
        )


@main.command('tags')
@click.argument('incomplete', metavar='PREFIX', required=True)
def tags_command(incomplete):
    """Get all tags starting with prefix.

    This command can be used in plugins to make tag autocompletion. Tags
    is separated by space. If tag starting with '-' or '--', use
    double-dash before it. Example:

    $ note tags -- --tag
    """

    async def wrapper():
        async with db_context() as db:
            return await complete_tag_name(db, incomplete, [])

    click.echo(' '.join(asyncio.run(wrapper())))


@main.command('view')
@click.argument('hex_id', metavar='ID', required=True)
def view_command(hex_id):
    """View note by id.

    Decorations in output are the same as in `search` command. Use
    `dump` if you want to get only note content.
    """

    async def wrapper():
        async with db_context() as db:
            await print_note_by_hex_id(db, hex_id)

    asyncio.run(wrapper())


if __name__ == '__main__':
    main()
