import asyncio
import math
import random
import sys
from hashlib import sha1

import click
from pygments import highlight, lexers
from pygments.formatters import (  # pylint: disable=no-name-in-module
    Terminal256Formatter,
)
from pygments.util import ClassNotFound

from note_storage.cli.errors import exit_on_unsuccesful_response
from note_storage.cli.utils import db_context
from note_storage.core.config import config
from note_storage.core.utils import chunks
from note_storage.gateways import SQLiteNoteGateway
from note_storage.interactors import GetNotesWithTagsByFieldInteractor
from note_storage.requests import GetNotesWithTagsByFieldRequest

MINIMAL_NOTE_LINES = 5


def get_border_color():
    return config.get('cli', 'border_color')


def pager_needed(n_lines):
    prompt_lines = 2
    _, terminal_lines = click.get_terminal_size()

    return n_lines > terminal_lines - prompt_lines


def intelligent_echo(text):
    n_lines = text.count('\n') + 1

    if pager_needed(n_lines):
        click.echo_via_pager(text)
    else:
        click.echo(text)


def color_by_hash(string):
    colors = ['red', 'green', 'yellow', 'blue', 'magenta', 'cyan']
    n_colors = len(colors)

    hex_hash = sha1(string.encode()).hexdigest()
    rand = int(hex_hash[:4], 16) / 65535
    index = math.floor(rand * n_colors)

    if index == n_colors:
        return colors[-1]

    return colors[math.floor(rand * n_colors)]


def find_nth(haystack, needle, n):
    # https://stackoverflow.com/a/1884277/12075481
    start = haystack.find(needle)

    while start >= 0 and n > 1:
        start = haystack.find(needle, start + len(needle))
        n -= 1

    return start


def trim_body_if_needed(note_body, max_body_lines):
    body_lines = note_body.count('\n') + 1

    if body_lines > max_body_lines:
        note_body = '\n'.join(
            [
                note_body[: find_nth(note_body, '\n', max_body_lines)],
                click.style('...', fg=get_border_color()),
            ]
        )

    return note_body


def highlight_syntax(body, extension):
    try:
        lexer = lexers.get_lexer_by_name(extension)
    except ClassNotFound:
        return body + '\n'

    return highlight(body, lexer, Terminal256Formatter(style='bw'))


def decorate_tags(tags, query):
    tag_names = sorted([tag.name for tag in tags])
    return ' '.join(
        [
            click.style(
                tag_name,
                bold=True,
                fg=color_by_hash(tag_name),
                underline=(tag_name in query),
            )
            for tag_name in tag_names
        ]
    )


def decorate_note_body(note, max_body_lines):
    note_body = (
        trim_body_if_needed(note.body, max_body_lines)
        if max_body_lines
        else note.body
    )

    return (
        highlight_syntax(note_body, note.extension)
        if note.extension != 'txt'
        else note_body + '\n'
    )


def decorate_note(note, tags, query=None, max_body_lines=0):
    if query is None:
        query = []

    decorated_tag_names = decorate_tags(tags, query)

    note_metadata = ''.join(
        [
            'ID: ',
            click.style(note.hex_id, bold=True),
            ' ' * 2,
            'Tags: ',
            decorated_tag_names,
        ]
    )

    note_body = decorate_note_body(note, max_body_lines)

    terminal_columns, _ = click.get_terminal_size()

    return ''.join(
        [
            click.style('-' * terminal_columns, fg=get_border_color()),
            '\n',
            note_metadata,
            '\n\n',
            note_body,
            click.style('─' * terminal_columns, fg=get_border_color()),
        ]
    )


async def decorate_notes_chunk(ids_chunk, query, shuffle, max_body_lines):
    async with db_context() as db:
        interactor = GetNotesWithTagsByFieldInteractor(SQLiteNoteGateway(db))

        response = await interactor.execute(
            await GetNotesWithTagsByFieldRequest.from_dict(
                {'field': 'id', 'values': ids_chunk}
            )
        )

        exit_on_unsuccesful_response(response)

        notes_with_tags = response.value
        # We need to shuffle again, 'cause "WHERE IN" flatted the order
        if shuffle:
            random.shuffle(notes_with_tags)

        return '\n'.join(
            [
                decorate_note(*note_with_tags, query, max_body_lines)
                for note_with_tags in response.value
            ]
        )


def print_notes_by_ids(note_ids, query, shuffle, max_body_lines):
    chunk_len = 32

    if len(note_ids) == 1:
        decorated_notes = asyncio.run(
            decorate_notes_chunk(note_ids, query, shuffle, max_body_lines=0)
        )
        intelligent_echo(decorated_notes)
        return

    if not pager_needed(len(note_ids) * MINIMAL_NOTE_LINES):
        decorated_notes = asyncio.run(
            decorate_notes_chunk(note_ids, query, shuffle, max_body_lines)
        )
        intelligent_echo(decorated_notes)
        return

    def decorate_notes_chunks_iterator(note_ids):
        for ids_chunk in chunks(note_ids, chunk_len):
            yield asyncio.run(
                decorate_notes_chunk(ids_chunk, query, shuffle, max_body_lines)
            )

    click.echo_via_pager(decorate_notes_chunks_iterator(note_ids))


async def print_note_by_hex_id(db, hex_id):
    interactor = GetNotesWithTagsByFieldInteractor(SQLiteNoteGateway(db))
    response = await interactor.execute(
        await GetNotesWithTagsByFieldRequest.from_dict(
            {'field': 'hex_id', 'values': [hex_id]}
        )
    )

    exit_on_unsuccesful_response(response)

    notes_with_tags = response.value

    if not notes_with_tags:
        sys.exit(f'Note "{hex_id}" not found')

    decorated_note = decorate_note(*(notes_with_tags[0]))
    intelligent_echo(decorated_note)
