import sys
from inspect import cleandoc

import click
from jsonschema import Draft7Validator

from note_storage.core.config.toml import TOMLConfig
from note_storage.core.jsonschema_validation import (
    synchronous_get_first_validation_error,
)
from note_storage.core.utils import (
    get_project_config_path,
    get_project_data_path,
)

config_validator = Draft7Validator(
    {
        'type': 'object',
        'properties': {
            'general': {
                'type': 'object',
                'properties': {
                    'data_path': {'type': 'string', 'pattern': r'^/'},
                    'shuffle_search': {'type': 'boolean'},
                    'default_extension': {'type': 'string'},
                },
                'additionalProperties': False,
            },
            'cli': {
                'type': 'object',
                'properties': {
                    'border_color': {
                        'type': 'string',
                        'enum': ['black', 'white'],
                    },
                    'instant_view': {'type': 'boolean'},
                },
                'additionalProperties': False,
            },
        },
        'additionalProperties': False,
    }
)


def get_default_config_text():
    return cleandoc(
        f'''
        [general]
        data_path = "{get_project_data_path()}"
        default_extension = "txt"
        shuffle_search = false

        [cli]
        border_color = "white"
        instant_view = false
        '''
    )


def ensure_config_exists(config_path):
    if not config_path.exists():
        config_path.parent.mkdir(exist_ok=True)
        config_path.write_text(get_default_config_text() + '\n')


def read_config(config_path):
    config_ = TOMLConfig()
    config_.loads(config_path.read_text())

    return config_


def validate_config(config_dict):
    if error := synchronous_get_first_validation_error(
        config_validator, config_dict
    ):
        error_message = f'Bad config file: {error[1]}'
        click.echo(error_message, err=True)
        sys.exit(1)


def fill_config_defaults(config_):
    if config_.get('general', 'data_path') is None:
        config_.set('general', 'data_path', str(get_project_data_path()))

    if config_.get('general', 'default_extension') is None:
        config_.set('general', 'default_extension', "txt")

    if config_.get('general', 'shuffle_search') is None:
        config_.set('general', 'shuffle_search', False)

    if config_.get('cli', 'border_color') is None:
        config_.set('cli', 'border_color', 'white')

    if config_.get('cli', 'instant_view') is None:
        config_.set('cli', 'instant_view', False)


def _get_config():
    config_path = get_project_config_path()

    ensure_config_exists(config_path)
    config_ = read_config(config_path)
    validate_config(config_.to_dict())
    fill_config_defaults(config_)

    return config_


config = _get_config()
