from itertools import chain

from note_storage.core.db.errors import SQLiteNotFoundError
from note_storage.core.db.utils import get_placeholders
from note_storage.domains import Note

from .tag import SQLiteTagGateway

LAST_REVIEWED_ID_NAME = "last_reviewed_id"


class SQLiteNoteGateway:
    def __init__(self, db):
        self.db = db

    @staticmethod
    def make_domain(note_data):
        return Note.from_dict(
            {
                'id': note_data['id'],
                'hex_id': note_data['hex_id'],
                'extension': note_data['extension'],
                'body': note_data['body'],
            }
        )

    async def get_all_notes_ids(self):
        async with self.db.execute(
            'SELECT id FROM notes ORDER BY id'
        ) as cursor:
            note_ids = await cursor.fetchall()

        return list(chain.from_iterable(note_ids))

    async def get_notes_with_tags_by_field(self, field, values):
        async with self.db.execute(
            'SELECT * FROM notes WHERE {} IN {}'.format(
                field, get_placeholders(len(values))
            ),
            (*values,),
        ) as cursor:
            notes_data = await cursor.fetchall()

        notes = [self.make_domain(note_data) for note_data in notes_data]
        tag_gateway = SQLiteTagGateway(self.db)

        return [
            (note, await tag_gateway.get_all_note_tags(note.id))
            for note in notes
        ]

    async def get_note_ids_by_tag_ids(self, tag_ids):
        tags_number = len(tag_ids)
        async with self.db.execute(
            '''
            SELECT note_id FROM note_tag
            WHERE tag_id IN {}
            GROUP BY note_id
            HAVING COUNT(*) = ?
            '''.format(
                get_placeholders(tags_number)
            ),
            (*tag_ids, tags_number),
        ) as cursor:
            note_ids = await cursor.fetchall()

        return list(chain.from_iterable(note_ids))

    async def get_hex_ids_by_note_ids(self, note_ids):
        async with self.db.execute(
            'SELECT hex_id FROM notes WHERE id IN {}'.format(
                get_placeholders(len(note_ids))
            ),
            (*note_ids,),
        ) as cursor:
            hex_ids = await cursor.fetchall()

        return list(chain.from_iterable(hex_ids))

    async def create_note(self, hex_id, ext, body, tag_names):
        async with self.db.execute(
            'INSERT INTO notes (hex_id, extension, body) VALUES (?, ?, ?)',
            (hex_id, ext, body),
        ) as cursor:
            note_id = cursor.lastrowid

        tag_gateway = SQLiteTagGateway(self.db)
        await self.link_tags_to_note(tag_gateway, note_id, set(tag_names))

        await self.db.commit()

    async def is_hex_id_available(self, hex_id):
        async with self.db.execute(
            'SELECT id FROM notes WHERE hex_id = ?', (hex_id,)
        ) as cursor:
            return (await cursor.fetchone()) is None

    async def remove_note(self, hex_id):
        async with self.db.execute(
            'DELETE FROM notes WHERE hex_id = ?', (hex_id,)
        ) as cursor:
            if not cursor.rowcount:
                raise SQLiteNotFoundError(f'Note "{hex_id}" not found')

        await self.db.commit()

        await SQLiteTagGateway(self.db).remove_orphaned_tags()

    async def unlink_tags_from_note(self, note_id, tag_names):
        await self.db.execute(
            '''
            DELETE FROM note_tag WHERE note_id = ? AND tag_id IN (
              SELECT id FROM tags WHERE name IN {}
            )
            '''.format(
                get_placeholders(len(tag_names))
            ),
            (note_id, *tag_names),
        )

        await self.db.commit()

    async def link_tags_to_note(self, tag_gateway, note_id, tag_names):
        for tag_name in tag_names:
            tag = await tag_gateway.get_or_create_tag(tag_name)

            await self.db.execute(
                'INSERT INTO note_tag (note_id, tag_id) VALUES (?, ?)',
                (note_id, tag.id),
            )

            await self.db.commit()

    async def update_note_record(self, note_id, note_ext, note_body):
        await self.db.execute(
            'UPDATE notes SET extension = ?, body = ? WHERE id = ?',
            (note_ext, note_body, note_id),
        )

        await self.db.commit()

    async def update_note(self, note_id, note_ext, note_body, tag_names):
        tag_gateway = SQLiteTagGateway(self.db)

        await self.update_note_record(note_id, note_ext, note_body)

        old_tag_names = {
            tag.name for tag in await tag_gateway.get_all_note_tags(note_id)
        }

        if removed_tag_names := old_tag_names - set(tag_names):
            await self.unlink_tags_from_note(note_id, removed_tag_names)

            await tag_gateway.remove_orphaned_tags()

        if new_tag_names := set(tag_names) - old_tag_names:
            await self.link_tags_to_note(tag_gateway, note_id, new_tag_names)

    async def get_last_reviewed_note_id(self):
        async with self.db.execute(
            'SELECT value FROM system_storage WHERE key = ?',
            (LAST_REVIEWED_ID_NAME,),
        ) as cursor:
            row = await cursor.fetchone()

        return int(row["value"]) if row else None

    async def set_last_reviewed_note_id(self, last_reviewed_id):
        await self.db.execute(
            '''
            INSERT INTO system_storage (key, value)
            VALUES (?, ?)
            ON CONFLICT(key)
            DO UPDATE SET value=excluded.value
            ''',
            (LAST_REVIEWED_ID_NAME, str(last_reviewed_id)),
        )

        await self.db.commit()

    async def delete_last_reviewed_note_id(self):
        await self.db.execute(
            'DELETE FROM system_storage WHERE key = ?',
            (LAST_REVIEWED_ID_NAME,),
        )

        await self.db.commit()
