import sqlite3
from itertools import chain

from note_storage.core.db.errors import (
    SQLiteNotFoundError,
    SQLiteUniqueConstraintError,
)
from note_storage.core.db.utils import get_placeholders
from note_storage.domains import Tag


class SQLiteTagGateway:
    def __init__(self, db):
        self.db = db

    @staticmethod
    def make_domain(tag_data):
        return Tag.from_dict({'id': tag_data['id'], 'name': tag_data['name']})

    async def get_tags_by(self, field, values):
        tags_data = []

        for value in values:
            cursor = await self.db.execute(
                f'SELECT * FROM tags WHERE {field}=?', (value,)
            )
            tag_data = await cursor.fetchone()
            await cursor.close()

            if tag_data is None:
                raise SQLiteNotFoundError(f'Tag "{value}" not found')

            tags_data.append(tag_data)

        return [self.make_domain(tag_data) for tag_data in tags_data]

    async def get_tag_completions(self, incomplete, completed_ids=None):
        if completed_ids and (completed_number := len(completed_ids)):
            await self.db.execute(
                '''
                CREATE TEMPORARY TABLE matched_note_ids AS
                  SELECT note_id FROM note_tag
                  WHERE tag_id IN {}
                  GROUP BY note_id
                  HAVING COUNT(*) = ?
                '''.format(
                    get_placeholders(completed_number)
                ),
                (*completed_ids, len(completed_ids)),
            )

            async with self.db.execute(
                '''
                SELECT name FROM (
                  SELECT id, name FROM tags WHERE name LIKE ? AND id IN (
                    SELECT tag_id FROM note_tag
                    WHERE note_id IN matched_note_ids AND tag_id NOT IN {}
                  )
                ) as completions
                INNER JOIN
                (
                  SELECT tag_id, count(*) AS n_links FROM note_tag
                  WHERE note_id in matched_note_ids
                  GROUP BY tag_id
                ) as nt
                WHERE completions.id = nt.tag_id
                ORDER BY n_links DESC, completions.name
                '''.format(
                    get_placeholders(completed_number)
                ),
                (incomplete + '%', *completed_ids),
            ) as cursor:
                completions_data = await cursor.fetchall()

            await self.db.execute('DROP TABLE matched_note_ids')
        else:
            async with self.db.execute(
                '''
                SELECT completion.name FROM (
                  SELECT * FROM tags WHERE name LIKE ?
                ) as completion
                INNER JOIN
                (
                  SELECT tag_id, count(*) AS n_links FROM note_tag
                  GROUP BY tag_id
                ) as nt
                WHERE completion.id = nt.tag_id
                ORDER BY n_links DESC, completion.name
                ''',
                (incomplete + '%',),
            ) as cursor:
                completions_data = await cursor.fetchall()

        return list(chain.from_iterable(completions_data))

    async def get_all_note_tags(self, note_id):
        async with self.db.execute(
            '''
            SELECT * FROM tags WHERE id IN (
              SELECT tag_id FROM note_tag WHERE note_id = ?
            )
            ''',
            (note_id,),
        ) as cursor:
            tags_data = await cursor.fetchall()

        return [self.make_domain(tag_data) for tag_data in tags_data]

    async def get_or_create_tag(self, name):
        async with self.db.execute(
            'INSERT OR IGNORE INTO tags (name) VALUES (?)', (name,)
        ) as cursor:
            if cursor.rowcount:
                await self.db.commit()

        async with self.db.execute(
            'SELECT * FROM tags WHERE name = ?', (name,)
        ) as cursor:
            tag = await cursor.fetchone()
            return self.make_domain({'id': tag['id'], 'name': tag['name']})

    async def remove_orphaned_tags(self):
        await self.db.execute(
            '''
            DELETE FROM tags WHERE id NOT IN (
              SELECT tag_id FROM note_tag
            )
            '''
        )

        await self.db.commit()

    async def rename_tag(self, old_name, new_name):
        try:
            await self.db.execute(
                'UPDATE tags SET name = ? WHERE name = ?', (new_name, old_name)
            )
        except sqlite3.IntegrityError as exc:
            if 'UNIQUE' in str(exc):
                raise SQLiteUniqueConstraintError(
                    f'Tag "{new_name}" already exists'
                ) from exc
            raise

        else:
            await self.db.commit()
